"""User.

UserModel for FTP Users.
"""
import uuid
import hashlib
import base64
import secrets
from asyncdb.models import Model, Column
from datetime import datetime


def at_now():
    return datetime.now()

class FTPUser(Model):
    userid: uuid.UUID = Column(required=False, primary_key=True, db_default='auto')
    username: str = Column(required=False)
    tenant: str = Column(required=False)
    password: str = Column(required=False)
    name: str = Column(required=False)
    last_login: datetime = Column(required=False, default=at_now)
    date_joined: datetime = Column(required=False)
    last_ip: str = Column(required=False)
    is_active: bool = Column(required=False, default=False)

    def validate_password(self, password) -> bool:
        try:
            algorithm, iterations, salt, hash = str(self.password).split("$", 3)
            assert algorithm == 'pbkdf2_sha256'
        except Exception:
            raise Exception(
                f'FTP Users: Invalid password Hash or scheme for user: {self.name}'
            )
        compare_hash = self.create_password(
            password,
            iterations=int(iterations),
            salt=salt,
            token_num=6
        )
        if secrets.compare_digest(self.password, compare_hash):
            return True
        return False

    def create_password(
        self,
        password: str,
        token_num: int = 6,
        iterations: int = 80000,
        salt: str = None
    ):
        if not salt:
            salt = secrets.token_hex(token_num)
        key = hashlib.pbkdf2_hmac(
            'sha256',
            password.encode("utf-8"),
            salt.encode("utf-8"),
            iterations,
            dklen=32,
        )
        strhash = base64.b64encode(key).decode("utf-8").strip()
        return f"pbkdf2_sha256${iterations}${salt}${strhash}"

    class Meta:
        driver = 'pg'
        name = 'ftpusers'
        schema = 'public'
        strict = True
        frozen = False



"""aioSFTP Meta information.
   FTP/SSH Servers implementation.
"""

__title__ = 'aiosftp'
__description__ = ('FTP/SSH/sFTP Server implementation built on to of Asyncio.'
                   'Facility to deploy SSH server easily inside any project.')
__version__ = '0.1.6'
__author__ = 'Jesus Lara'
__author_email__ = 'jesuslara@phenoarbital.info'
__license__ = 'Apache-2'

def get_version() -> tuple: # pragma: no cover
   """
   Get aiosftp server version as a tuple.
   """
   return tuple(x for x in __version__.split('.')) # pragma: no cover



#!/usr/bin/env python
"""aioSFTP.

    Create an FTP Server for upload files with integrated watchdogs and security.
See:
https://github.com/phenobarbital/aiosftp
"""
import ast
from os import path
from setuptools import find_packages, setup


def get_path(filename):
    return path.join(path.dirname(path.abspath(__file__)), filename)


def readme():
    with open(get_path('README.md'), 'r', encoding='utf-8') as rd:
        return rd.read()


version = get_path('aiosftp/version.py')
with open(version, 'r', encoding='utf-8') as meta:
    t = compile(meta.read(), version, 'exec', ast.PyCF_ONLY_AST)
    for node in (n for n in t.body if isinstance(n, ast.Assign)):
        if len(node.targets) == 1:
            name = node.targets[0]
            if isinstance(name, ast.Name) and \
                name.id in (
                    '__version__',
                    '__title__',
                    '__description__',
                    '__author__',
                    '__license__', '__author_email__'):
                v = node.value
                if name.id == '__version__':
                    __version__ = v.s
                if name.id == '__title__':
                    __title__ = v.s
                if name.id == '__description__':
                    __description__ = v.s
                if name.id == '__license__':
                    __license__ = v.s
                if name.id == '__author__':
                    __author__ = v.s
                if name.id == '__author_email__':
                    __author_email__ = v.s

setup(
    name=__title__,
    version=__version__,
    python_requires=">=3.8.0",
    url='https://github.com/phenobarbital/aiosftp',
    description=__description__,
    long_description=readme(),
    long_description_content_type='text/markdown',
    keywords = "ftp, ftp server, sftp server, ssh server, asyncio",
    license=__license__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Object Brokering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Networking',
        'Operating System :: OS Independent',
        'Environment :: Web Environment',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Framework :: AsyncIO'
    ],
    author='Jesus Lara',
    author_email='jesuslara@phenobarbital.info',
    packages=find_packages(exclude=["contrib", "docs", "tests", "settings"]),
    setup_requires=[
        "asyncio==3.4.3",
    ],
    install_requires=[
        "aiohttp==3.8.3",
        "asyncio==3.4.3",
        "uvloop==0.17.0",
        "aiofiles==0.8.0",
        "aiofile==3.8.1",
        "aioftp==0.21.4",
        "siosocks>=0.3.0",
        "asyncssh==2.12.0",
        "asyncdb[default]>=2.1.30",
        "navconfig>=1.0.6"
    ],
    tests_require=[
            'pytest>=5.4.0',
            'coverage',
            'pytest-asyncio==0.20.1',
            'pytest-xdist==3.0.2',
            'pytest-assume==2.4.2'
    ],
    entry_points={
        'console_scripts': [
            'aioftp = aiosftp.__main__:main'
        ]
    },
    project_urls={  # Optional
        'Source': 'https://github.com/phenobarbital/aiosftp',
        'Funding': 'https://paypal.me/phenobarbital',
        'Say Thanks!': 'https://saythanks.io/to/phenobarbital',
    },
)



"""
aioSFTP Configuration.
"""
# Import Config Class
from pathlib import Path
from navconfig import config, BASE_DIR

"""
Databases
"""
# DB Default
# POSTGRESQL Default
DBHOST = config.get('DBHOST', fallback='localhost')
DBUSER = config.get('DBUSER')
DBPWD = config.get('DBPWD')
DBNAME = config.get('DBNAME', fallback='navigator')
DBPORT = config.get('DBPORT', fallback=5432)
if not DBUSER:
    raise Exception('Missing PostgreSQL Default Settings.')

# database for changes (admin)
default_dsn = 'postgres://{user}:{password}@{host}:{port}/{db}'.format(
    user=DBUSER,
    password=DBPWD,
    host=DBHOST,
    port=DBPORT,
    db=DBNAME
)


"""
FTP Server
"""
FTP_ENABLED = config.getboolean('ENABLE_FTP', fallback=False)
# main config for FTP:
FTP_SERVER_HOST = config.get('FTP_SERVER_HOST', fallback='127.0.0.1')
FTP_SERVER_PORT = config.getint('FTP_SERVER_PORT', fallback=2122)

"""
SSH Server
"""
SSH_ENABLED = config.getboolean('ENABLE_SSH', fallback=False)

SSH_SERVER_HOST = config.get('SSH_SERVER_HOST', fallback='127.0.0.1')
SSH_SERVER_PORT = config.getint('SSH_SERVER_PORT', fallback=8384)

## path for serving files
SERVICE_BASE_PATH = Path(config.get('SERVICE_BASE_PATH', fallback=BASE_DIR))

### Secure Connections
SSL_CERT = config.get('SFTP_SSL_CERT', fallback='docs/ssl/ssl-cert-snakeoil.pem')
SSL_KEY = config.get('SFTP_SSL_KEY', fallback='docs/ssl/ssl-cert-snakeoil.key')
CERT_CHECK_HOSTNAME = config.get('SFTP_SSL_CHECK_HOSTNAME', fallback=False)

ECDSA_KEY = config.get('ECDSA_KEY', fallback=BASE_DIR.joinpath('docs', 'ssl', 'ssh_host_ecdsa_key'))
RSA_KEY = config.get('RSA_KEY', fallback=BASE_DIR.joinpath('docs', 'ssl', 'ssh_host_rsa_key'))

### SSH Configuration
SSHD_CONFIG = config.get('SSHD_CONFIG', fallback=BASE_DIR.joinpath('docs', 'ssl', 'sshd_config'))
SSH_KNOWN_HOSTS = config.get('SSH_KNOWN_HOSTS', fallback=BASE_DIR.joinpath('docs', 'ssl', 'known_hosts'))

try:
    from settings.settings import * # pylint: disable=W0614,W0401
except ImportError as ex:
    print(ex)



"""FTP
FTP Server implemented in Asyncio.
AioFTP Server.
"""
import os
import ssl
import pathlib
import socket
import asyncio
import uvloop
import aioftp
from aioftp.server import AbstractUserManager, AvailableConnections
from navconfig import BASE_DIR
from navconfig.logging import logging
from .conf import (
    FTP_SERVER_HOST,
    FTP_SERVER_PORT,
    SERVICE_BASE_PATH,
    SSL_CERT,
    SSL_KEY,
    CERT_CHECK_HOSTNAME
)

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

try:
    DEFAULT_HOST = FTP_SERVER_HOST
except Exception as e:
    DEFAULT_HOST = socket.gethostbyname(socket.gethostname())


class UserModel(AbstractUserManager):
    path: pathlib.Path = None

    def __init__(self, path, *args, **kwargs):
        self.path = path
        super(UserModel, self).__init__(timeout=10, *args, **kwargs)
        self.available_connections = {}

    async def get_user(self, login: str) -> aioftp.User:
        user = None
        print(login)
        perm = []
        state = None
        info = ''
        perm.append(
            aioftp.Permission('/', readable=True, writable=True)
        )
        # perm.append(
        #     aioftp.Permission(home_path, readable=True, writable=True)
        # )
        user = aioftp.User(
            login=login,
            password=login,
            base_path=self.path,
            home_path=pathlib.PurePosixPath("/"),
            maximum_connections=2,
            permissions=perm
        )
        print(user)
        self.available_connections[user] = AvailableConnections(
            user.maximum_connections)
        if user is None:
            state = AbstractUserManager.GetUserResponse.ERROR
            info = "no such username"
        if self.available_connections[user].locked():
            state = AbstractUserManager.GetUserResponse.ERROR
            info = f"too much connections for {user.login or 'anonymous'!r}"
        elif user.login is None:
            state = AbstractUserManager.GetUserResponse.ERROR
            info = "Anonymous login prohibited"
        elif user.password is None:
            state = AbstractUserManager.GetUserResponse.PASSWORD_REQUIRED
            info = "Password required"
        else:
            state = AbstractUserManager.GetUserResponse.OK
            info = 'User OK'
        print(state, user, info)
        # if state != AbstractUserManager.GetUserResponse.ERROR:
        #     self.available_connections[user].acquire()
        return state, user, info

    async def authenticate(self, user: aioftp.User, password: str) -> bool:
        print(user, password)
        return True

    async def notify_logout(self, user: aioftp.User):
        print(user)
        self.available_connections[user].release()
        return user


class FTPServer(object):
    """aio FTP server.

    Attributes:
        host: Hostname of the server.
        port: Port number of the server.
        loop: Event loop to run in.
        kwargs: are passed directly to aioftp server.
    """
    _server: aioftp.Server = None

    def __init__(
            self,
            host: str = DEFAULT_HOST,
            port: int = FTP_SERVER_PORT,
            path: str = SERVICE_BASE_PATH,
            event_loop: asyncio.AbstractEventLoop = None,
            debug: bool = False,
            users: list = None,
            **kwargs
    ):
        self.host = host
        self.port = port
        self.path = path
        self.debug = debug
        self.log = logging.getLogger("aio.ftp")
        self.ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        self.ssl_ctx.load_cert_chain(
            certfile=str(
                BASE_DIR.joinpath(SSL_CERT)
            ),
            keyfile=str(
                BASE_DIR.joinpath(SSL_KEY)
            )
        )
        self.ssl_ctx.options |= ssl.OP_NO_TLSv1
        self.ssl_ctx.options |= ssl.OP_NO_TLSv1_1
        self.ssl_ctx.options |= ssl.OP_SINGLE_DH_USE
        self.ssl_ctx.options |= ssl.OP_SINGLE_ECDH_USE
        self.ssl_ctx.check_hostname = CERT_CHECK_HOSTNAME
        self.ssl_ctx.verify_mode = ssl.VerifyMode.CERT_NONE
        self.ssl_ctx.set_ciphers(
            'ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384')

        if self.debug:
            logging.basicConfig(
                level=logging.DEBUG,
                format="%(asctime)s [%(name)s] %(message)s",
                datefmt="[%H:%M:%S]:",
            )
            self.log.setLevel(logging.DEBUG)
        if not event_loop:
            self._loop = asyncio.get_event_loop()
            asyncio.set_event_loop(self._loop)
        else:
            self._loop = event_loop
        try:
            os.chdir(str(self.path))
            print(f"FTP: Working Directory: {os.getcwd()}")
            users = UserModel(self.path)
            self._server = aioftp.Server(
                users=users,
                path_io_factory=aioftp.AsyncPathIO,
                socket_timeout=30,
                idle_timeout=60,
                path_timeout=10,
                encoding='utf-8',
                maximum_connections=10,
                ssl=self.ssl_ctx
            )
        except Exception as err:
            print(err)
            raise

    async def close(self):
        """Close.
        Closing FTP server connections
        """
        try:
            await self._server.close()
            self.log.info('::: aio-FTP: Closing all FTP connections.')
        except Exception as err:
            logging.exception(err)

    async def start(self):
        """Starts server."""
        # Serve requests until Ctrl+C is pressed
        self.log.info(
            f'::: aio-FTP: Serving FTP Server {self.host} on {self.port}'
        )
        # running forever
        await self._server.start(
            host=self.host,
            port=int(self.port),
            family=socket.AF_INET,
            loop=self._loop
        )



"""SCP

SCP/sFTP Server implemented in Asyncio.
Asyncssh Server.
"""
import os
import ssl
import socket
import asyncio
from datetime import datetime
import asyncssh
from navconfig.logging import logging
from asyncdb import AsyncDB
from .conf import (
    default_dsn,
    SSH_SERVER_HOST,
    SSH_SERVER_PORT,
    SERVICE_BASE_PATH,
    SSL_CERT,
    SSL_KEY,
    CERT_CHECK_HOSTNAME,
    ECDSA_KEY,
    RSA_KEY,
    SSH_KNOWN_HOSTS,
    SSHD_CONFIG,
)
from .user import FTPUser

ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
ctx.load_cert_chain(
    certfile=SSL_CERT,
    keyfile=SSL_KEY
)
ctx.options |= ssl.OP_NO_TLSv1
ctx.options |= ssl.OP_NO_TLSv1_1
ctx.options |= ssl.OP_SINGLE_DH_USE
ctx.options |= ssl.OP_SINGLE_ECDH_USE
ctx.check_hostname = CERT_CHECK_HOSTNAME
ctx.set_ciphers('ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384')

asyncssh.set_debug_level(1)
asyncssh.set_log_level(logging.INFO)
logging.basicConfig(level='INFO')

try:
    DEFAULT_HOST = SSH_SERVER_HOST
except Exception as e:
    DEFAULT_HOST = socket.gethostbyname(socket.gethostname())


class aioSSHServer(asyncssh.SSHServer):
    def connection_made(self, conn):
        self.connection = conn
        self.ip = conn.get_extra_info('peername')[0]
        print(f'{self.ip} - Connection received')

    def connection_lost(self, exc):
        print(f'{self.ip} - Connection closed {exc}')
        if exc is not None:
            logging.exception(exc)

    async def begin_auth(self, username: str) -> bool:
        """Handle client authentication request"""
        logging.debug(
            f'sFTP: Starting authentication for user: {username}'
        )
        return True

    def password_auth_supported(self):
        """Return whether or not password authentication is supported."""
        return True

    def kbdint_auth_supported(self):
        return False

    def public_key_auth_supported(self):
        """Return whether or not public key authentication is supported"""
        return False

    async def validate_password(self, username, password):
        try:
            db = AsyncDB('pg', dsn=default_dsn)
            async with await db.connection() as conn:
                FTPUser.Meta.set_connection(conn)
                user = await FTPUser.get(username=username)
                if user.validate_password(password):
                    try:
                        user.last_ip = self.ip
                        user.last_login = datetime.now()
                        await user.update()
                        self.connection.set_extra_info(user=user)
                    except Exception as err:
                        logging.error(err)
                    return True
                else:
                    return False
        except Exception as err:
            logging.exception(
                f"sFTP: Cannot connect with user Subsystem {err}"
            )
            return False

    def auth_completed(self, **kwargs) -> None:
        """Authentication was completed successfully."""
        return True


class aioSFTPServer(asyncssh.SFTPServer):
    def __init__(self, chan: asyncssh.SSHServerChannel):
        root = SERVICE_BASE_PATH
        try:
            user = chan.get_extra_info('user')
            tenant = user.tenant
            if tenant:
                root = SERVICE_BASE_PATH.joinpath(tenant)
        except Exception as err:
            logging.error(err)
        if not root.exists():
            # create automatically
            root.mkdir(exist_ok=True, parents=True)
        logging.debug(
            f'Connecting to Directory: {root}'
        )
        super().__init__(chan, chroot=str(root))


class SCPServer(object):
    """aio sFTP server.

    Attributes:
        host: Hostname of the server.
        port: Port number of the server.
        loop: Event loop to run in.
        kwargs: are passed directly to asyncssh server.
    """
    _server: asyncssh.SFTPServer = None

    def __init__(
            self,
            host: str = DEFAULT_HOST,
            port: int = SSH_SERVER_PORT,
            path: str = SERVICE_BASE_PATH,
            event_loop: asyncio.AbstractEventLoop = None,
            debug: bool = False,
            **kwargs
    ):
        self.host = host
        self.port = port
        self.path = path
        self.debug = debug
        if self.debug is True:
            asyncssh.set_debug_level(1)
            asyncssh.set_log_level(logging.DEBUG)
        self.log = logging.getLogger("aio.ssh")
        self.ssh_key = ECDSA_KEY
        if not self.ssh_key.exists():
            self.ssh_key = RSA_KEY
        self.ssh_config = SSHD_CONFIG
        self.ssh_host_keys = asyncssh.read_private_key_list(self.ssh_key)
        self._known_hosts = SSH_KNOWN_HOSTS
        if self.debug:
            dbg = logging.DEBUG
            asyncssh.set_log_level(dbg)
            asyncssh.set_debug_level(1)
            asyncssh.set_sftp_log_level(1)
            logging.basicConfig(
                level=dbg,
                format="%(asctime)s [%(name)s] %(message)s",
                datefmt="[%H:%M:%S]:",
            )
            self.log.setLevel(dbg)
        if not event_loop:
            self._loop = asyncio.get_event_loop()
            asyncio.set_event_loop(self._loop)
        else:
            self._loop = event_loop
        try:
            os.chdir(str(self.path))
            print(f"SSH: Working Directory: {os.getcwd()}")
        except Exception as err:
            print(err)
            raise

    async def close(self):
        """Close.
        Closing FTP server connections
        """
        try:
            self.log.info('::: aio-sFTP: Closing all sFTP connections.')
            self._server.close()
        except Exception as err:
            logging.exception(err)

    def handle_clients(self, process: asyncssh.SSHServerProcess) -> None:
        user = process.get_extra_info('username')
        process.stdout.write(f'Hi {user}, Welcome to aioSFTP SSH server\n')

    async def start(self):
        """Starts server."""
        # Serve requests until Ctrl+C is pressed
        self.log.info(
            f'::: aio-sFTP: Serving sFTP Server {self.host} on {self.port}'
        )
        # running forever
        self._server = await asyncssh.listen(
            server_factory=aioSSHServer,
            sftp_factory=aioSFTPServer,
            process_factory=self.handle_clients,
            host=self.host,
            port=int(self.port),
            server_host_keys=[self.ssh_key],
            family=socket.AF_INET,
            allow_scp=True,
            password_auth=True,
            host_based_auth=True,
            known_client_hosts=[self._known_hosts],
            config=str(self.ssh_config)
        )
        # print('sFTP Server: ', self._server)



"""aio (s)FTP server entry point."""
import asyncio
import argparse
import uvloop
from .ftp import FTPServer
from .scp import SCPServer
from .conf import (
    FTP_SERVER_HOST,
    FTP_SERVER_PORT,
    SSH_SERVER_HOST,
    SSH_SERVER_PORT,
    SERVICE_BASE_PATH
)


asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
uvloop.install()
loop = asyncio.new_event_loop()
asyncio.set_event_loop(loop)


def main():
    """Main Worker Function."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument('--host', dest='host', type=str,
                        default=SSH_SERVER_HOST,
                        help='set server host'
                        )
    parser.add_argument('--port', dest='port', type=int,
                        default=SSH_SERVER_PORT,
                        help='set server port'
                        )
    parser.add_argument('--ftp-host', dest='ftp_host', type=str,
                        default=FTP_SERVER_HOST,
                        help='Set server FTP Host'
                        )
    parser.add_argument('--ftp-port', dest='ftp_port', type=int,
                        default=FTP_SERVER_PORT,
                        help='Set server FTP port'
                        )
    parser.add_argument('--path', dest='path', type=str,
                        default=SERVICE_BASE_PATH,
                        help='Base path for FTP and SSH clients'
                        )
    parser.add_argument('--debug', action="store_true", default=False,
                        help="Start (s)FTP server in Debug Mode"
                        )
    args = vars(parser.parse_args())
    if not args['ftp_host']:
        args['ftp_host'] = args['host']
    print('::: Starting FTP and sFTP Server')
    ftp_args = {
        "host": args['ftp_host'],
        "port": args['ftp_port'],
        "path": args['path'],
        "debug": args['debug'],
    }
    try:
        ftp = FTPServer(**ftp_args, event_loop=loop)
        scp = SCPServer(**args, event_loop=loop)
        loop.run_until_complete(
            ftp.start()
        )
        loop.run_until_complete(
            scp.start()
        )
        loop.run_forever()
    except Exception as e:
        raise
    except KeyboardInterrupt:
        loop.run_until_complete(ftp.close())
        loop.run_until_complete(scp.close())
        print('Closing (s)FTP Connections ...')
    finally:
        loop.close()


if __name__ == '__main__':
    main()
